/*
 * Image streaming
 *
 * Copyright IBM, Corp. 2011
 *
 * Authors:
 *  Stefan Hajnoczi   <stefanha@linux.vnet.ibm.com>
 *
 * This work is licensed under the terms of the GNU LGPL, version 2 or later.
 * See the COPYING.LIB file in the top-level directory.
 *
 */

#include "qemu/osdep.h"
#include "trace.h"
#include "block/block_int.h"
#include "block/blockjob_int.h"
#include "qapi/error.h"
#include "qapi/qmp/qerror.h"
#include "qapi/qmp/qdict.h"
#include "qemu/ratelimit.h"
#include "sysemu/block-backend.h"
#include "block/copy-on-read.h"
#include "block/aio_task.h"

enum {
    /*
     * Maximum chunk size to feed to copy-on-read.  This should be
     * large enough to process multiple clusters in a single call, so
     * that populating contiguous regions of the image is efficient.
     */
    STREAM_CHUNK = 512 * 1024, /* in bytes */
};

#define STREAM_CLUSTER_SIZE_DEFAULT (64 * 1024)

/* Maximum number of stream aio tasks in flight */
#define STREAM_MAX_IN_FLIGHT 16

typedef struct StreamAioTask StreamAioTask;
typedef struct StreamBlockJob {
    BlockJob common;
    BlockDriverState *base_overlay; /* COW overlay (stream from this) */
    BlockDriverState *above_base;   /* Node directly above the base */
    BlockDriverState *cor_filter_bs;
    BlockDriverState *target_bs;
    BlockdevOnError on_error;
    char *backing_file_str;
    bool bs_read_only;
    uint64_t len;
    int64_t cluster_size;
    int64_t max_chunk;

    /* Cached result of last call to is_allocated on bs->backing */
    int64_t backing_allocated_offset;
    int64_t backing_allocated_end;

    AioTaskPool *aio;
    QSIMPLEQ_HEAD(, StreamAioTask) retry_tasks;
    int ignored_ret;
} StreamBlockJob;

struct StreamAioTask {
    AioTask task;

    StreamBlockJob *stream_job;
    QSIMPLEQ_ENTRY(StreamAioTask) entry;
    int64_t offset;
    int64_t bytes;
};

static int coroutine_fn stream_task(AioTask *task)
{
    StreamAioTask *t = container_of(task, StreamAioTask, task);
    StreamBlockJob *s = t->stream_job;
    int ret;

    assert(t->bytes < SIZE_MAX);

    trace_stream_read(s, t->offset, t->bytes,
                      aio_task_pool_busy_tasks(task->pool));
    ret = blk_co_preadv(s->common.blk, t->offset, t->bytes, NULL,
                        BDRV_REQ_PREFETCH);
    trace_stream_read_finish(s, t->offset, t->bytes, ret,
                             aio_task_pool_busy_tasks(task->pool));
    if (ret < 0) {
        BlockErrorAction action =
                block_job_error_action(&s->common, s->on_error, true, -ret);

        switch (action) {
        case BLOCK_ERROR_ACTION_REPORT:
            return ret;
        case BLOCK_ERROR_ACTION_STOP:
            t = g_memdup(t, sizeof(*t));
            QSIMPLEQ_INSERT_HEAD(&s->retry_tasks, t, entry);
            return 0;
        case BLOCK_ERROR_ACTION_IGNORE:
            if (s->ignored_ret == 0) {
                s->ignored_ret = ret;
            }
            break;
        default:
            abort();
        }
    }

    /* BLOCK_ERROR_ACTION_IGNORE or success */

    job_progress_update(&s->common.job, t->bytes);

    return 0;
}

static void coroutine_fn start_stream_task(StreamBlockJob *s,
                                           int64_t offset, int64_t bytes)
{
    StreamAioTask *task = g_new(StreamAioTask, 1);

    *task = (StreamAioTask) {
        .task.func = stream_task,
        .stream_job = s,
        .offset = offset,
        .bytes = bytes,
    };

    aio_task_pool_start_task(s->aio, &task->task);
}

static int stream_prepare(Job *job)
{
    StreamBlockJob *s = container_of(job, StreamBlockJob, common.job);
    BlockDriverState *unfiltered_bs = bdrv_skip_filters(s->target_bs);
    BlockDriverState *base;
    BlockDriverState *unfiltered_base;
    Error *local_err = NULL;
    int ret = 0;

    /* We should drop filter at this point, as filter hold the backing chain */
    bdrv_cor_filter_drop(s->cor_filter_bs);
    s->cor_filter_bs = NULL;

    base = bdrv_filter_or_cow_bs(s->above_base);
    unfiltered_base = bdrv_skip_filters(base);

    if (bdrv_cow_child(unfiltered_bs)) {
        const char *base_id = NULL, *base_fmt = NULL;
        if (unfiltered_base) {
            base_id = s->backing_file_str ?: unfiltered_base->filename;
            if (unfiltered_base->drv) {
                base_fmt = unfiltered_base->drv->format_name;
            }
        }
        bdrv_set_backing_hd(unfiltered_bs, base, &local_err);
        ret = bdrv_change_backing_file(unfiltered_bs, base_id, base_fmt, false);
        if (local_err) {
            error_report_err(local_err);
            return -EPERM;
        }
    }

    return ret;
}

static void stream_clean(Job *job)
{
    StreamBlockJob *s = container_of(job, StreamBlockJob, common.job);
    BlockJob *bjob = &s->common;

    if (s->cor_filter_bs) {
        bdrv_cor_filter_drop(s->cor_filter_bs);
        s->cor_filter_bs = NULL;
    }

    /* Reopen the image back in read-only mode if necessary */
    if (s->bs_read_only) {
        /* Give up write permissions before making it read-only */
        blk_set_perm(bjob->blk, 0, BLK_PERM_ALL, &error_abort);
        bdrv_reopen_set_read_only(s->target_bs, true, NULL);
    }

    g_free(s->backing_file_str);
}

/*
 * Check if some bytes starting from offset are allocated in backing chain
 * (base..backing(bs)] with use of the cache.
 *
 * We can cache result for backing layers as they are unchanged during stream
 * operation.
 *
 * We need to cache only allocated ranges, as we just skip unallocated ones by
 * one jump.
 */
static int is_allocated_in_source_layers(StreamBlockJob *s,
                                         int64_t offset, int64_t *n)
{
    int ret;
    BlockDriverState *bs = blk_bs(s->common.blk);
    BlockDriverState *unfiltered_bs = bdrv_skip_filters(bs);

    if (offset >= s->backing_allocated_offset &&
        offset < s->backing_allocated_end)
    {
        *n = s->backing_allocated_end - offset;
        return 1;
    }

    ret = bdrv_is_allocated_above(bdrv_cow_bs(unfiltered_bs), s->base_overlay,
                                  true, offset, INT_MAX, n);
    if (ret == 1) {
        s->backing_allocated_offset = offset;
        s->backing_allocated_end = offset + *n;
    }

    return ret;
}

/*
 * stream_block_status
 *
 * Process some bytes starting from offset, return n aligned to cluster_size
 * (except for end of disk), and copy flag, which defines what to do with these
 * n bytes starting from offset: copy or skip.
 */
static int stream_block_status(StreamBlockJob *s, int64_t offset,
                               int64_t *n, bool *copy)
{
    int ret;
    int64_t remaining = s->len - offset;
    BlockDriverState *bs = blk_bs(s->common.blk);
    BlockDriverState *unfiltered_bs = bdrv_skip_filters(bs);
    int64_t n2;

    assert(QEMU_IS_ALIGNED(offset, s->cluster_size));
    *copy = false;

    ret = bdrv_is_allocated(unfiltered_bs, offset, s->max_chunk, n);
    if (ret < 0) {
        return ret;
    }

    if (ret == 1) {
        /*
         * Allocated in the top, no need to copy. Try to skip as much as
         * possible.
         *
         * Note, that we don't want to use INT_MAX in first call of
         * bdrv_is_allocated, as in unallocated case, most of returned
         * information may be unuseful, and we can't cache it, as top bs is
         * changing.
         */
        ret = bdrv_is_allocated(bs, offset + *n, INT_MAX, &n2);
        if (ret == 1) {
            *n += n2;
        }
        goto skip_n_bytes;
    }

    /*
     * The area on top bs is unallocated, we are going to copy (if it is
     * allocated in underlying levels above base). Align n up.
     */
    *n = MIN(remaining, QEMU_ALIGN_UP(*n, s->cluster_size));

    ret = is_allocated_in_source_layers(s, offset, &n2);
    if (ret < 0) {
        return ret;
    }

    if (ret > 0) {
        *copy = true;
        /*
         * n  - the number of bytes unallocated in the top layer, aligned up
         *      except for EOF
         * n2 - the number of bytes allocated in source layers at the given
         *      offset.
         */
        *n = MIN(*n, QEMU_ALIGN_UP(n2, s->cluster_size));
    } else if (ret == 0 && n2 == 0) {
        /* Finish early if end of backing file has been reached */
        *n = s->len - offset;
    } else {
        /* Skip n2 bytes unallocated in backing */
        *n = n2;
        goto skip_n_bytes;
    }

    return 0;

skip_n_bytes:
    assert(!*copy && *n > 0);
    if (*n < remaining) {
        if (*n < s->cluster_size) {
            *copy = true;
            *n = MIN(remaining, s->cluster_size);
        } else {
            *n = QEMU_ALIGN_DOWN(*n, s->cluster_size);
        }
    } else if (*n > remaining) {
        *n = remaining;
    }

    return 0;
}

static int coroutine_fn stream_run(Job *job, Error **errp)
{
    StreamBlockJob *s = container_of(job, StreamBlockJob, common.job);
    BlockDriverState *unfiltered_bs = bdrv_skip_filters(s->target_bs);
    int64_t offset = 0;
    uint64_t delay_ns = 0;
    int64_t n = 0; /* bytes */
    int ret = 0;

    if (unfiltered_bs == s->base_overlay) {
        /* Nothing to stream */
        return 0;
    }

    job_progress_set_remaining(&s->common.job, s->len);

    s->aio = aio_task_pool_new(STREAM_MAX_IN_FLIGHT);

    for ( ; offset < s->len; offset += n) {
        bool copy;

        /* Note that even when no rate limit is applied we need to yield
         * with no pending I/O here so that bdrv_drain_all() returns.
         */
        job_sleep_ns(&s->common.job, delay_ns);
        if (job_is_cancelled(&s->common.job) || aio_task_pool_status(s->aio)) {
            break;
        }

        ret = stream_block_status(s, offset, &n, &copy);
        if (ret < 0) {
            BlockErrorAction action =
                block_job_error_action(&s->common, s->on_error, true, -ret);
            if (action == BLOCK_ERROR_ACTION_STOP) {
                n = 0;
                continue;
            }
            if (s->ignored_ret == 0) {
                s->ignored_ret = ret;
            }
            if (action == BLOCK_ERROR_ACTION_REPORT) {
                break;
            }
        }
        trace_stream_one_iteration(s, offset, n, copy);

        if (copy) {
            start_stream_task(s, offset, n);
            delay_ns = block_job_ratelimit_get_delay(&s->common, n);
        } else {
            job_progress_update(&s->common.job, n);
            delay_ns = 0;
        }
    }

    /*
     * We will want to handle retries if in-flight requests fail
     * with BLOCK_ERROR_ACTION_STOP
     */
    while ((!aio_task_pool_empty(s->aio) || !QSIMPLEQ_EMPTY(&s->retry_tasks))
           && !job_is_cancelled(&s->common.job))
    {
        aio_task_pool_wait_all(s->aio);
        job_pause_point(&s->common.job);
    }

    aio_task_pool_wait_all(s->aio);
    while (!QSIMPLEQ_EMPTY(&s->retry_tasks)) {
        StreamAioTask *task = QSIMPLEQ_FIRST(&s->retry_tasks);
        QSIMPLEQ_REMOVE_HEAD(&s->retry_tasks, entry);
        g_free(task);
    }
    if (ret == 0) {
        ret = aio_task_pool_status(s->aio);
    }
    g_free(s->aio);
    s->aio = NULL;

    /* Do not remove the backing file if an error was there but ignored. */
    return ret ?: s->ignored_ret;
}

static void stream_pause(Job *job)
{
    StreamBlockJob *s = container_of(job, StreamBlockJob, common.job);

    aio_task_pool_wait_all(s->aio);
}

static void stream_resume(Job *job)
{
    StreamBlockJob *s = container_of(job, StreamBlockJob, common.job);

    while (!QSIMPLEQ_EMPTY(&s->retry_tasks)) {
        StreamAioTask *task = QSIMPLEQ_FIRST(&s->retry_tasks);
        QSIMPLEQ_REMOVE_HEAD(&s->retry_tasks, entry);

        aio_task_pool_start_task(s->aio, &task->task);
    }
}

static const BlockJobDriver stream_job_driver = {
    .job_driver = {
        .instance_size = sizeof(StreamBlockJob),
        .job_type      = JOB_TYPE_STREAM,
        .free          = block_job_free,
        .run           = stream_run,
        .prepare       = stream_prepare,
        .clean         = stream_clean,
        .user_resume   = block_job_user_resume,
        .pause         = stream_pause,
        .resume        = stream_resume,
    },
};

void stream_start(const char *job_id, BlockDriverState *bs,
                  BlockDriverState *base, const char *backing_file_str,
                  BlockDriverState *bottom,
                  int creation_flags, int64_t speed,
                  BlockdevOnError on_error,
                  const char *filter_node_name,
                  Error **errp)
{
    StreamBlockJob *s = NULL;
    BlockDriverState *iter;
    bool bs_read_only;
    int basic_flags = BLK_PERM_CONSISTENT_READ | BLK_PERM_WRITE_UNCHANGED;
    BlockDriverState *base_overlay;
    BlockDriverState *cor_filter_bs = NULL;
    BlockDriverState *above_base;
    QDict *opts;
    int ret;
    int64_t len = bdrv_getlength(bs);
    int64_t cluster_size;
    BlockDriverInfo bdi;

    if (len < 0) {
        error_setg_errno(errp, -len, "unable to get length for '%s'",
                         bdrv_get_device_or_node_name(bs));
        return;
    }

    assert(!(base && bottom));
    assert(!(backing_file_str && bottom));

    if (bottom) {
        /*
         * New simple interface. The code is written in terms of old interface
         * with @base parameter (still, it doesn't freeze link to base, so in
         * this mean old code is correct for new interface). So, for now, just
         * emulate base_overlay and above_base. Still, when old interface
         * finally removed, we should refactor code to use only "bottom", but
         * not "*base*" things.
         */
        assert(!bottom->drv->is_filter);
        base_overlay = above_base = bottom;
    } else {
        base_overlay = bdrv_find_overlay(bs, base);
        if (!base_overlay) {
            error_setg(errp, "'%s' is not in the backing chain of '%s'",
                       base->node_name, bs->node_name);
            return;
        }

        /*
         * Find the node directly above @base.  @base_overlay is a COW overlay,
         * so it must have a bdrv_cow_child(), but it is the immediate overlay
         * of @base, so between the two there can only be filters.
         */
        above_base = base_overlay;
        if (bdrv_cow_bs(above_base) != base) {
            above_base = bdrv_cow_bs(above_base);
            while (bdrv_filter_bs(above_base) != base) {
                above_base = bdrv_filter_bs(above_base);
            }
        }
    }

    ret = bdrv_get_info(bs, &bdi);
    cluster_size = ret < 0 ? STREAM_CLUSTER_SIZE_DEFAULT : bdi.cluster_size;

    /* Make sure that the image is opened in read-write mode */
    bs_read_only = bdrv_is_read_only(bs);
    if (bs_read_only) {
        int ret;
        /* Hold the chain during reopen */
        if (bdrv_freeze_backing_chain(bs, above_base, errp) < 0) {
            return;
        }

        ret = bdrv_reopen_set_read_only(bs, false, errp);

        /* failure, or cor-filter will hold the chain */
        bdrv_unfreeze_backing_chain(bs, above_base);

        if (ret < 0) {
            return;
        }
    }

    opts = qdict_new();

    qdict_put_str(opts, "driver", "copy-on-read");
    qdict_put_str(opts, "file", bdrv_get_node_name(bs));
    /* Pass the base_overlay node name as 'bottom' to COR driver */
    qdict_put_str(opts, "bottom", base_overlay->node_name);
    if (filter_node_name) {
        qdict_put_str(opts, "node-name", filter_node_name);
    }

    cor_filter_bs = bdrv_insert_node(bs, opts, BDRV_O_RDWR, errp);
    if (!cor_filter_bs) {
        goto fail;
    }

    if (!filter_node_name) {
        cor_filter_bs->implicit = true;
    }

    s = block_job_create(job_id, &stream_job_driver, NULL, cor_filter_bs,
                         BLK_PERM_CONSISTENT_READ,
                         basic_flags | BLK_PERM_WRITE,
                         speed, creation_flags, NULL, NULL, errp);
    if (!s) {
        goto fail;
    }

    /*
     * Prevent concurrent jobs trying to modify the graph structure here, we
     * already have our own plans. Also don't allow resize as the image size is
     * queried only at the job start and then cached.
     */
    if (block_job_add_bdrv(&s->common, "active node", bs, 0,
                           basic_flags | BLK_PERM_WRITE, errp)) {
        goto fail;
    }

    /* Block all intermediate nodes between bs and base, because they will
     * disappear from the chain after this operation. The streaming job reads
     * every block only once, assuming that it doesn't change, so forbid writes
     * and resizes. Reassign the base node pointer because the backing BS of the
     * bottom node might change after the call to bdrv_reopen_set_read_only()
     * due to parallel block jobs running.
     * above_base node might change after the call to
     * bdrv_reopen_set_read_only() due to parallel block jobs running.
     */
    base = bdrv_filter_or_cow_bs(above_base);
    for (iter = bdrv_filter_or_cow_bs(bs); iter != base;
         iter = bdrv_filter_or_cow_bs(iter))
    {
        ret = block_job_add_bdrv(&s->common, "intermediate node", iter, 0,
                                 basic_flags, errp);
        if (ret < 0) {
            goto fail;
        }
    }

    s->base_overlay = base_overlay;
    s->above_base = above_base;
    s->backing_file_str = g_strdup(backing_file_str);
    s->cor_filter_bs = cor_filter_bs;
    s->target_bs = bs;
    s->bs_read_only = bs_read_only;
    s->len = len;
    s->cluster_size = cluster_size;
    s->max_chunk = MAX(cluster_size, STREAM_CHUNK);

    s->on_error = on_error;
    trace_stream_start(bs, base, s);
    job_start(&s->common.job);
    return;

fail:
    if (s) {
        job_early_fail(&s->common.job);
    }
    if (cor_filter_bs) {
        bdrv_cor_filter_drop(cor_filter_bs);
    }
    if (bs_read_only) {
        bdrv_reopen_set_read_only(bs, true, NULL);
    }
}
