/*
 * VZ Kernel bitmaps
 *
 * Copyright (c) 2022 Virtuozzo International GmbH
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "qemu/osdep.h"

#include <linux/fs.h>
#include <sys/ioctl.h>

#include "qapi/error.h"
#include "qemu/uuid.h"
#include "block/block.h"
#include "block/vz-kernel-cbt.h"

/* Copied from block/blk-cbt.c in kernel source */
#define CBT_MAX_EXTENTS  512

static int vz_cbt_start(int fd, QemuUUID *uuid, Error **errp)
{
    int ret;
    struct blk_user_cbt_info ci = {
        .ci_blksize = 64 * 1024,
    };

    memcpy(ci.ci_uuid, uuid->data, 16);

    ret = ioctl(fd, BLKCBTSTART, &ci);
    if (ret < 0) {
        error_setg_errno(errp, errno, "ioctl BLKCBTSTART failed");
        return -errno;
    }

    return 0;
}

static int vz_cbt_merge_back(int fd, QemuUUID *uuid, Error **errp)
{
    int ret;
    struct blk_user_cbt_misc_info cmi = {
        .action = CBT_SNAP_MERGE_BACK,
    };

    memcpy(cmi.uuid, uuid->data, 16);

    ret = ioctl(fd, BLKCBTMISC, &cmi);
    if (errno == ENODEV) {
        /* It's OK: no snapshot, nothing to merge */
        return 0;
    }
    if (ret < 0) {
        error_setg_errno(errp, errno,
                         "ioctl BLKCBTMISC(CBT_SNAP_MERGE_BACK) failed");
        return -errno;
    }

    return 0;
}

/*
 * @ci is OUT-only parameter, vz_cbt_get() resets all fields
 * @ci->extents array must be enough to store CBT_MAX_EXTENTS extents
 */
static int vz_cbt_get(int fd, uint64_t offset, uint64_t bytes,
                      const QemuUUID *uuid, const char *uuidstr,
                      struct blk_user_cbt_info *ci, Error **errp)
{
    int ret;
    int i;
    uint64_t prev_extent_end = 0;
    int64_t end = offset + bytes;

    *ci = (struct blk_user_cbt_info) {
        .ci_extent_count = CBT_MAX_EXTENTS,
        .ci_start = offset,
        .ci_length = bytes,
    };
    ret = ioctl(fd, BLKCBTGET, ci);
    if (ret < 0) {
        error_setg_errno(errp, errno, "ioctl BLKCBTGET failed");
        return -errno;
    }

    if (memcmp(uuid->data, ci->ci_uuid, 16)) {
        char unexpected[UUID_FMT_LEN + 1];

        qemu_uuid_unparse((void *)ci->ci_uuid, unexpected);
        error_setg(errp, "Kernel reply to BLKCBTGET with unexpected uuid: "
                   "%s, expected: %s", unexpected, uuidstr);
        return -EINVAL;
    }

    for (i = 0; i < ci->ci_mapped_extents; i++) {
        struct blk_user_cbt_extent *e = &ci->ci_extents[i];
        uint64_t extent_end = e->ce_physical + e->ce_length;

        if (e->ce_physical < offset || extent_end > end) {
            error_setg(errp, "Kernel reply to BLKCBTGET(offset = %" PRIu64
                       ", bytes = %" PRIu64 ") contains out-of-bound extent"
                       " (offset = %llu, bytes = %llu)",
                       offset, bytes, e->ce_physical, e->ce_length);
            return -EINVAL;
        }

        if (i > 0 && prev_extent_end > e->ce_physical) {
            error_setg(errp, "Kernel reply to BLKCBTGET contains "
                       "intersecting or unordered extents");
            return -EINVAL;
        }

        prev_extent_end = extent_end;
    }

    return 0;
}

static struct blk_user_cbt_info *vz_cbt_new_info(void)
{
    return g_malloc0(sizeof(struct blk_user_cbt_info) +
                     CBT_MAX_EXTENTS * sizeof(struct blk_user_cbt_extent));
}

int vz_cbt_pull(int fd, const char *uuidstr,
                BdrvDirtyBitmap *target, Error **errp)
{
    int ret;
    g_autofree struct blk_user_cbt_info *ci = vz_cbt_new_info();
    int64_t offset = 0;
    int64_t bytes = bdrv_dirty_bitmap_size(target);
    QemuUUID uuid;

    qemu_uuid_parse(uuidstr, &uuid);

    ret = vz_cbt_merge_back(fd, &uuid, errp);
    if (ret < 0) {
        return ret;
    }

    while (bytes) {
        int i;
        uint64_t done;
        struct blk_user_cbt_extent *last_e;

        ret = vz_cbt_get(fd, offset, bytes, &uuid, uuidstr, ci, errp);
        if (ret < 0) {
            return ret;
        }

        for (i = 0; i < ci->ci_mapped_extents; i++) {
            bdrv_set_dirty_bitmap(target, ci->ci_extents[i].ce_physical,
                                  ci->ci_extents[i].ce_length);
        }

        if (ci->ci_mapped_extents < CBT_MAX_EXTENTS) {
            /* No more extents I hope */
            break;
        }

        last_e = &ci->ci_extents[ci->ci_mapped_extents - 1];
        done = last_e->ce_physical + last_e->ce_length - offset;
        bytes -= done;
        offset += done;
    }

    /* BLKCBTSTOP is not critical, ignore failure */
    ioctl(fd, BLKCBTSTOP, NULL);

    return 0;
}

/*
 * @ci is OUT-only parameter. Function sets
 *   @ci->ci_extent_count
 *   @ci->ci_mapped_extents
 *   @ci->extents
 *
 * @ci->extents array must be enough to store CBT_MAX_EXTENTS extents
 *
 * Other @ci fields are untouched.
 */
static void vz_cbt_bitmap_to_extents(BdrvDirtyBitmap *bitmap,
                                     uint64_t offset, uint64_t bytes,
                                     struct blk_user_cbt_info *ci)
{
    int64_t start, dirty_start, dirty_count;
    int64_t end = offset + bytes;

    bdrv_dirty_bitmap_lock(bitmap);

    ci->ci_extent_count = 0;
    for (start = offset;
         ci->ci_extent_count < CBT_MAX_EXTENTS &&
         bdrv_dirty_bitmap_next_dirty_area(bitmap, start, end, INT32_MAX,
                                           &dirty_start, &dirty_count);
         start = dirty_start + dirty_count)
    {
        ci->ci_extents[ci->ci_extent_count].ce_physical = dirty_start;
        ci->ci_extents[ci->ci_extent_count].ce_length = dirty_count;
        ci->ci_extent_count++;
    }

    ci->ci_mapped_extents = ci->ci_extent_count;

    bdrv_dirty_bitmap_unlock(bitmap);
}

int vz_cbt_push(int fd, const char *uuidstr,
                BdrvDirtyBitmap *source, Error **errp)
{
    int ret;
    g_autofree struct blk_user_cbt_info *ci = vz_cbt_new_info();
    int64_t offset = 0;
    int64_t bytes = bdrv_dirty_bitmap_size(source);
    QemuUUID uuid;

    qemu_uuid_parse(uuidstr, &uuid);

    ret = vz_cbt_start(fd, &uuid, errp);
    if (ret < 0) {
        return ret;
    }

    memcpy(ci->ci_uuid, uuid.data, 16);

    while (bytes) {
        uint64_t done;
        struct blk_user_cbt_extent *last_e;

        vz_cbt_bitmap_to_extents(source, offset, bytes, ci);
        if (ci->ci_mapped_extents == 0) {
            /* No more extents */
            break;
        }

        ret = ioctl(fd, BLKCBTSET, ci);
        if (ret < 0) {
            error_setg_errno(errp, errno, "ioctl BLKCBTSET failed");
            return -errno;
        }

        if (ci->ci_mapped_extents < CBT_MAX_EXTENTS) {
            /* No more extents */
            break;
        }

        last_e = &ci->ci_extents[ci->ci_mapped_extents - 1];
        done = last_e->ce_physical + last_e->ce_length - offset;
        bytes -= done;
        offset += done;
    }

    return 0;
}
